#!/usr/bin/perl -wT
# $Id: flowdump.pl,v 0.0.1 2002/03/21 12:00:57 raven Exp $
require 5.6.0;
use strict;
use warnings;
use English;
use Carp qw(croak);
use Sys::Hostname qw();
use Socket qw(:DEFAULT);
use Getopt::Long qw();
use Pod::Usage qw();
use Net::Netmask qw();
use POSIX qw(strftime);
use Time::HiRes qw(gettimeofday);
use vars qw($packet %packet %service %protocol);
$| = 1;

# Netflow v.5 packet format
%packet = (
	bufsize => 1464,
	hdrfmt => 'nnNNNNNA*',
	statfmt => 'a4a4a4nnNNNNnnCCCCnnCCnA*'
);

sub get_bind_iface() {
  my $name = Sys::Hostname::hostname();
  return (gethostbyname($name))[4];
}

sub get_options() {
  my %default = (
	port => 901,
	bind => get_bind_iface()
  );
  Getopt::Long::Configure(qw(
	no_ignore_case
	require_order
	no_permute
  ));
  Getopt::Long::GetOptions(\my %opt,
	'sa=s',
 	'da=s',
	'sp=s',
	'dp=s',
	'ss=i',
	'ds=i',
	'match=s',
	'p|proto=s',
	'bind=s',
	'port=i',
	'h|help!'
  );
  Pod::Usage::pod2usage(-verbose => 2) if $opt{'h'};
  if ($opt{'p'}) {
    if ($opt{'p'} =~ /^\d+$/) {
      getprotobynumber($opt{'p'}) or delete $opt{'p'};
    } else {
      $opt{'p'} = (getprotobyname($opt{'p'}))[2] or delete $opt{'p'};
    }
    croak 'Wrong protocol specified' unless $opt{'p'};
  }
  foreach (qw(sp dp)) {
    next unless $opt{$_};
    my @port;
    (@port) = $opt{$_} =~ /^(\d+)(:\d+)?$/;
    if (@port) {
      substr($port[1], 0, 1, '') if $port[1];
      foreach (@port) {
        next unless $_;
        croak 'Port exceeds range' if $_ < 0 or $_ > 65535;
      }
    } else {
      croak 'No protocol specified' unless $opt{'p'};
      croak 'Wrong protocol' unless $opt{'p'} == 6 or $opt{'p'} == 17;
      $port[0] = (getservbyname($opt{$_}, getprotobynumber($opt{'p'})))[2];
      croak 'Wrong service' unless $port[0];
    }
    $port[1] = $port[0] unless $port[1];
    undef $opt{$_};
    @{$opt{$_}}[0,1] = ($port[1] >= $port[0] ? @port[0,1] : @port[1,0]);
  }
  foreach (keys %default) {
    $opt{$_} ||= $default{$_};
    ($opt{$_}) = $opt{$_} =~ /^(.*)$/;
  }
  foreach (qw(match sa da)) {
    $opt{$_} ? ($opt{$_} = new2 Net::Netmask ($opt{$_})) : next;
    croak 'Wrong address specification' unless $opt{$_};
    if ($opt{'match'}) {
      $opt{'sa'} = $opt{'da'} = $opt{'match'};
      last;
    }
  }
  return %opt;
}

sub bind_socket($) {
  my $opt = shift;
  socket(SOCK, PF_INET, SOCK_DGRAM, 17) or croak "socket(): $!";
  my $socket = pack('Sna4x8', AF_INET, @$opt{qw(port bind)});
  croak "sockaddr_in(): $!" unless $socket;
  bind(SOCK, $socket) or croak "bind(): $!";
}

sub close_socket() {
  my $signal = shift;
  close(SOCK);
  croak sprintf "Got SIG%s", $signal if $signal;
  exit;
}

sub match_ipaddr($$) {
  my $addr = shift or return();
  my $block = shift or return(1);
  return($block->match($addr) ? 1 : undef);
}

sub match_port($$) {
  my $port = shift or return();
  my $match = shift or return(1);
  return(($port >= $$match[0] and $port <= $$match[1]) ? 1 : undef);
}

sub match_as_or_proto($$) {
  my $val = shift;
  my $val_to_match = shift or return(1);
  return (($val and $val == $val_to_match) ? 1 : undef);
}

sub parse_flow($@) {
  my $opt = shift;
  my @flows = @_ or return();
  foreach my $flow (@flows[1..$#flows]) {
    my @data = unpack($packet{'statfmt'}, $flow);
    $data[$_] = inet_ntoa($data[$_]) foreach (0..2);
    unless ($protocol{$data[13]}) {
      $protocol{$data[13]} = (getprotobynumber($data[13]))[0] || $data[13];
    }
    foreach (9,10) {
      next if exists $service{$data[13]}{$data[$_]};
      $service{$data[13]}{$data[$_]} =
	getservbyport($data[$_], $protocol{$data[13]}) || $data[$_];
    }
    if ($$opt{'match'}) {
      next unless (
	match_ipaddr($data[0], $$opt{'sa'}) or
	match_ipaddr($data[1], $$opt{'da'})
      );
    } else {
      next unless (
	match_ipaddr($data[0], $$opt{'sa'}) and
	match_ipaddr($data[1], $$opt{'da'})
      );
    }
    next unless match_port($data[9], $$opt{'sp'});
    next unless match_port($data[10], $$opt{'dp'});
    next unless match_as_or_proto($data[15], $$opt{'ss'});
    next unless match_as_or_proto($data[16], $$opt{'ds'});
    next unless match_as_or_proto($data[13], $$opt{'p'});
    printf "%s.%s %s.%s > %s.%s: %s %s\n",
	strftime("%H:%M:%S", localtime(time())),
	(gettimeofday)[1],
	$data[0],
	$service{$data[13]}{$data[9]},
	$data[1],
	$service{$data[13]}{$data[10]},
	$protocol{$data[13]},
	$data[6];
  }
}

sub split_flow($) {
  my $opt = shift;
  recv(SOCK, my $packet, $packet{'bufsize'}, 0);
  return() unless $packet;
  my @flows = substr($packet, 0, 24, '');
  push(@flows, substr($packet, 0 , 48, '')) while $packet;
  parse_flow($opt,@flows);
}

$SIG{$_} = \&close_socket foreach qw(INT HUP USR1 USR2);
my %opt = get_options();
bind_socket(\%opt);
split_flow(\%opt) while (1);
close_socket();

__DATA__

=head1 NAME

flowdump.pl - simple tcpdump analog for netflow data parsing.

=head1 SYNOPSIS

flowdump.pl [ options ]

=head1 OPTIONS

=over 8

=item B<-sa SOURCE>

Source block of adresses in CIDR notation.

=item B<-da DESTINATION>

Destination block of adresses in CIDR notation. If both source AND
destination blocks are defined, 'AND' operator is used to match the
packet (See also --match option).

=item B<--match BLOCK>

Source OR destination block of adresses in CIDR notation.

=item B<-sp PORT[:PORT]>

Source port(-s). Port can be defined as service name (e.g. 'http'); this
requires --proto option.

=item B<-dp PORT[:PORT]>

Destination port(-s) or service.

=item B<-ss AS>

Source AS number.

=item B<-ds AS>

Destination AS number.

=item B<--proto PROTOCOL>

Protocol to match. Can be defined as number (e.g. tcp - 17) or as name.

=item B<--bind ADDRESS>

Interface address to bind. Unless defined, hostname resolution is performed.

=item B<--port PORT>

Port to bind. Default is 901.

=item B<-h|--help>

This help message.

=back

=head1 DESCRIPTION

B<flowdump.pl> is a simple tcpdump analog, that uses netflow data from cisco
equipment. It is used mostly for debugging or security audit.

=head1 AUTHOR

Vladimir Podgorny <raven@cemetery.org.ua>, Thursday, 21 Mar 2002.

Special thanks to Tony S. Hariman <tonyh@noc.cbn.net.id>.

=cut
